using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;           // for XmlTextReader and XmlValidatingReader
using System.Xml.Schema;    // for XmlSchemaCollection
using System.Diagnostics;   // for Debug
using System.IO;            // for File
using System.Configuration; // for ConfigurationManager

namespace QD_GUI
{
    public class XmlQueueDump
    {
        // If a validation error occurs,
        // set this flag to false in the
        // validation event handler. 
        private static bool isValid = true;
        private static string[] saHasChildElements = { 
            "queue_dump",
            "qd",
            "queue",
            "index",
            "server",
            "genome",
            "folding",
            "core",
            "cpu",
            "os",
            "assignment_info",
            "work_unit",
            "download_rate",
            "upload_rate",
            "performance_fraction"
        };

        // Display any validation errors.
        private static void ValidationCallBack(object sender, ValidationEventArgs e)
        {
            isValid = false;
            Debug.WriteLine("Validation Error: {0}", e.Message);
        }

        private static bool HasChildElements(string vsElement)
        {         
            foreach(string sParent in saHasChildElements)
            {
                if(vsElement == sParent) 
                {
                    return true;
                }
            }

            return false;
        }
        
        public static QueueDump ParseXmlFile(string vsXmlFile, QueueDump roQueueDump)
        {
            //////////////////////////////////////////////
            // Read the file with an XML reader

            if (File.Exists(vsXmlFile))
            {
                #region XML DTD

                // Elements in the qd XML ouput:
                //
                // Legend: + one or more occurences
                //         * zero or more occurences
                //         ? zero or one occurence(s)
                //
                // <!DOCTYPE queue_dump [
                //  <!ELEMENT queue_dump                (
                //                                       qd,
                //                                       qdinfo?,
                //                                       notice*,
                //                                       error?,
                //                                       queue?
                //                                      )>
                //  <!ELEMENT qd                        (
                //                                       date,
                //                                       functional_revision
                //                                      )>
                //  <!ELEMENT date                      (#PCDATA)>
                //  <!ELEMENT functional_revision       (#PCDATA)>
                //  <!ELEMENT qdinfo                    (#PCDATA)>
                //  <!ELEMENT notice                    (#PCDATA)>
                //  <!ELEMENT error                     (#PCDATA)>
                //  <!ELEMENT queue                     (
                //                                       version,
                //                                       system_type?,
                //                                       warning?,
                //                                       current_index,
                //                                       index+,
                //                                       results_successfully_sent?,
                //                                       download_rate?,
                //                                       upload_rate?,
                //                                       performance_fraction?,
                //                                       pph?,
                //                                       ppd?,
                //                                       ppw?,
                //                                       ppy?
                //                                      )>
                //  <!ELEMENT version                   (#PCDATA)>
                //  <!ELEMENT system_type               (#PCDATA)>
                //  <!ELEMENT warning                   (#PCDATA)>
                //  <!ELEMENT current_index             (#PCDATA)>
                //  <!ELEMENT index                     (
                //                                       number,
                //                                       status,
                //                                       points?,
                //                                       points_per_hour?,
                //                                       progress_rate?,
                //                                       progress?,
                //                                       server*,
                //                                       project_number?,
                //                                       project_name?,
                //                                       genome*,
                //                                       folding*,
                //                                       issue_timestamp?,
                //                                       begin_timestamp?,
                //                                       end_timestamp?,
                //                                       expect_timestamp?,
                //                                       due_timestamp?,
                //                                       expire_timestamp?,
                //                                       deadline?,
                //                                       preferred_timestamp?,
                //                                       preferred_deadline?,
                //                                       core?,
                //                                       cpu?,
                //                                       os?,
                //                                       smp_cores?,
                //                                       tag?,
                //                                       passkey?,
                //                                       memory?,
                //                                       assignment_info?,
                //                                       collection_server?,
                //                                       download_timestamp?,
                //                                       upload_failures?,
                //                                       packet_size_limit?,
                //                                       qgen?,
                //                                       username?,
                //                                       team?,
                //                                       id?,
                //                                       machine_id?,
                //                                       work_unit?
                //                                      )>
                //  <!ELEMENT number                    (#PCDATA)>
                //  <!ELEMENT status                    (#PCDATA)>
                //  <!ELEMENT points                    (#PCDATA)>
                //  <!ELEMENT points_per_hour           (#PCDATA)>
                //  <!ELEMENT progress_rate             (#PCDATA)>
                //  <!ELEMENT progress                  (#PCDATA)>
                //  <!ELEMENT server                    (
                //                                       number,
                //                                       address
                //                                      )>
                //  <!ELEMENT address                   (#PCDATA)>
                //  <!ELEMENT project_number            (#PCDATA)>
                //  <!ELEMENT project_name              (#PCDATA)>
                //  <!ELEMENT genome                    (
                //                                       unit?,
                //                                       benchmark?,
                //                                       misc?,
                //                                       other?
                //                                      )>
                //  <!ELEMENT unit                      (#PCDATA)>
                //  <!ELEMENT benchmark                 (#PCDATA)>
                //  <!ELEMENT misc                      (#PCDATA)>
                //  <!ELEMENT other                     (#PCDATA)>
                //  <!ELEMENT folding                   (
                //                                       run?,
                //                                       clone?,
                //                                       generation?,
                //                                       benchmark?,
                //                                       misc?,
                //                                       other?
                //                                      )>
                //  <!ELEMENT run                       (#PCDATA)>
                //  <!ELEMENT clone                     (#PCDATA)>
                //  <!ELEMENT generation                (#PCDATA)>
                //  <!ELEMENT issue_timestamp           (#PCDATA)>
                //  <!ELEMENT begin_timestamp           (#PCDATA)>
                //  <!ELEMENT end_timestamp             (#PCDATA)>
                //  <!ELEMENT expect_timestamp          (#PCDATA)>
                //  <!ELEMENT due_timestamp             (#PCDATA)>
                //  <!ELEMENT expire_timestamp          (#PCDATA)>
                //  <!ELEMENT deadline                  (#PCDATA)>
                //  <!ELEMENT preferred_timestamp       (#PCDATA)>
                //  <!ELEMENT preferred_deadline        (#PCDATA)>
                //  <!ELEMENT core                      (
                //                                       url?,
                //                                       version?
                //                                      )>
                //  <!ELEMENT url                       (#PCDATA)>
                //  <!ELEMENT cpu                       (
                //                                       cosm_type?,
                //                                       cosm_subtype?,
                //                                       name?
                //                                      )>
                //  <!ELEMENT cosm_type                 (#PCDATA)>
                //  <!ELEMENT cosm_subtype              (#PCDATA)>
                //  <!ELEMENT name                      (#PCDATA)>
                //  <!ELEMENT os                        (
                //                                       cosm_type?,
                //                                       cosm_subtype?,
                //                                       name?
                //                                      )>
                //  <!ELEMENT smp_cores                 (#PCDATA)>
                //  <!ELEMENT tag                       (#PCDATA)>
                //  <!ELEMENT passkey                   (#PCDATA)>
                //  <!ELEMENT memory                    (#PCDATA)>
                //  <!ELEMENT assignment_info           (
                //                                       byte_order?,
                //                                       timestamp?,
                //                                       checksum?
                //                                      )>
                //  <!ELEMENT byte_order                (#PCDATA)>
                //  <!ELEMENT timestamp                 (#PCDATA)>
                //  <!ELEMENT checksum                  (#PCDATA)>
                //  <!ELEMENT collection_server         (#PCDATA)>
                //  <!ELEMENT download_timestamp        (#PCDATA)>
                //  <!ELEMENT upload_failures           (#PCDATA)>
                //  <!ELEMENT packet_size_limit         (#PCDATA)>
                //  <!ELEMENT qgen                      (#PCDATA)>
                //  <!ELEMENT username                  (#PCDATA)>
                //  <!ELEMENT team                      (#PCDATA)>
                //  <!ELEMENT id                        (#PCDATA)>
                //  <!ELEMENT machine_id                (#PCDATA)>
                //  <!ELEMENT work_unit                 (
                //                                       file,
                //                                       size,
                //                                       type?
                //                                      )>
                //  <!ELEMENT file                      (#PCDATA)>
                //  <!ELEMENT size                      (#PCDATA)>
                //  <!ELEMENT type                      (#PCDATA)>
                //  <!ELEMENT results_successfully_sent (#PCDATA)>
                //  <!ELEMENT download_rate             (
                //                                       speed,
                //                                       units
                //                                      )>
                //  <!ELEMENT speed                     (#PCDATA)>
                //  <!ELEMENT units                     (#PCDATA)>
                //  <!ELEMENT upload_rate               (
                //                                       speed,
                //                                       units
                //                                      )>
                //  <!ELEMENT performance_fraction      (
                //                                       fraction,
                //                                       units
                //                                      )>
                //  <!ELEMENT fraction                  (#PCDATA)>
                //  <!ELEMENT pph                       (#PCDATA)>
                //  <!ELEMENT ppd                       (#PCDATA)>
                //  <!ELEMENT ppw                       (#PCDATA)>
                //  <!ELEMENT ppy                       (#PCDATA)>
                //]>

                #endregion XML DTD

                XmlReaderSettings oXmlReaderSettings = new XmlReaderSettings();
                oXmlReaderSettings.ProhibitDtd = false;
                oXmlReaderSettings.ValidationType = ValidationType.DTD;
                oXmlReaderSettings.ValidationEventHandler += new ValidationEventHandler(ValidationCallBack);

                XmlTextReader oXmlTextReader = new XmlTextReader(vsXmlFile);
                XmlReader oXmlReader = XmlReader.Create(oXmlTextReader, oXmlReaderSettings);

                return ParseXml(oXmlReader, roQueueDump);
            }
            else
            {
                throw new Exception("XML file does not exist!");
            }
        }

        public static QueueDump ParseXmlStream(Stream voXmlStream, QueueDump roQueueDump)
        {
            XmlReaderSettings oXmlReaderSettings = new XmlReaderSettings();
            oXmlReaderSettings.ProhibitDtd = false;
            oXmlReaderSettings.ValidationType = ValidationType.DTD;
            oXmlReaderSettings.ValidationEventHandler += new ValidationEventHandler(ValidationCallBack);

            XmlTextReader oXmlTextReader = new XmlTextReader(voXmlStream);
            XmlReader oXmlReader = XmlReader.Create(oXmlTextReader, oXmlReaderSettings);

            return ParseXml(oXmlReader, roQueueDump);
        }

        public static QueueDump ParseXml(XmlReader voXmlReader, QueueDump roQueueDump) 
        {
            #region Read XML

            string sParentElement = "";
            string sCurrentElement = "";
            int iCurrentIndex = -1;
            int iCurrentServer = 0;

            while (voXmlReader.Read())
            {
                // Get the node type
                XmlNodeType oXmlNodeType = voXmlReader.NodeType;

                switch (oXmlNodeType)
                {
                    // if node type is an element
                    case XmlNodeType.Element:
                        Debug.WriteLine("Element:" + voXmlReader.Name.ToString());

                        if(HasChildElements(sCurrentElement))
                        {
                            sParentElement = sCurrentElement;
                        }
                        sCurrentElement = voXmlReader.Name;

                        if (sCurrentElement == "index")
                        {
                            // processing a new index tag
                            // increment the index number
                            iCurrentIndex++;

                            // reset the Server index
                            iCurrentServer = 0;

                            //  Indices         
                            for (int i = 0; i < 10; i++)
                            {
                                if (roQueueDump.Queue.Index[i] == null)
                                {
                                    roQueueDump.Queue.Index[i] = new Index();

                                    // initialize all null referencing servers for this index
                                    for (int s = 0; s < 2; s++)
                                    {
                                        roQueueDump.Queue.Index[i].Server[s] = new Server();
                                    }
                                }
                            }
                        }
                        break;
                    case XmlNodeType.Text:
                        Debug.WriteLine("Text:" + voXmlReader.Value.ToString());

                        switch (sCurrentElement)
                        {
                            // Elements in the qd XML ouput:
                            // Legend: + one or more occurences
                            //         * zero or more occurences
                            //         ? zero or one occurence(s)
                            //  queue_dump
                            //   qd
                            //    date
                            case "date":
                                // set the property of the qd object
                                roQueueDump.QD.Date = voXmlReader.Value;
                                break;
                            //    functional_revision
                            case "functional_revision":
                                // set the property of the qd object
                                roQueueDump.QD.FunctionalRevision = voXmlReader.Value;

                                // Check if the qd functional revision used is at least 047
                                // The in 047 two tag where renamed:
                                //  credit          -> points
                                //  credit_per_hour -> points_per_hour
                                int iFunctionalRevision = Convert.ToInt32(voXmlReader.Value);
                                int iFunctionalRevisionRequired = Convert.ToInt32(Properties.Settings.Default.FunctionalRevisionRequired);
                                if (iFunctionalRevision < iFunctionalRevisionRequired) 
                                {
                                    string sDownloadUrl = Properties.Settings.Default.DownloadUrl;

                                    StringBuilder oString = new StringBuilder();
                                    oString.AppendFormat("Unsupported functional revision: {0}\r\n", voXmlReader.Value);
                                    oString.AppendFormat("\r\nFunctional Revision should be at least: {0}.\r\n", Properties.Settings.Default.FunctionalRevisionRequired);
                                    oString.AppendFormat("\r\nDownload the latest qd.exe at: {0}\r\n", sDownloadUrl);

                                    roQueueDump.Error = oString.ToString();
                                }

                                break;
                            //   qdinfo?
                            case "qdinfo":
                                // set the property
                                roQueueDump.QDInfo = QDInfo.ParseQDInfoString(voXmlReader.Value);
                                break;
                            //   notice*
                            case "notice":
                                // set the property
                                roQueueDump.Notice.Add(voXmlReader.Value);
                                break;
                            //   error?
                            case "error":
                                // set the property
                                roQueueDump.Error = voXmlReader.Value;
                                break;
                            //   queue?
                            //    version
                            case "version":
                                // version is used as an attribute by these tags:
                                //  queue
                                //  core
                                if (sParentElement == "queue")
                                {
                                    roQueueDump.Queue.Version = voXmlReader.Value;
                                }
                                else if (sParentElement == "core")
                                {
                                    roQueueDump.Queue.Index[iCurrentIndex].Core.Version = voXmlReader.Value;
                                }
                                break;
                            //    system_type?
                            case "system_type":
                                // set the property
                                roQueueDump.Queue.SystemType = voXmlReader.Value;
                                break;
                            //    warning?
                            case "warning":
                                // set the property
                                roQueueDump.Queue.Warning = voXmlReader.Value;
                                break;
                            //    current_index
                            case "current_index":
                                // set the property
                                roQueueDump.Queue.CurrentIndex = voXmlReader.Value;
                                break;
                            //    index+
                            //     number
                            case "number":
                                // use by the tags:
                                //  index
                                //  server
                                if (sParentElement == "index")
                                {
                                    int iIndexNumber = Convert.ToInt32(voXmlReader.Value);
                                    iCurrentIndex = iIndexNumber;

                                    // set the property
                                    roQueueDump.Queue.Index[iCurrentIndex].Number = voXmlReader.Value;
                                }
                                else if (sParentElement == "server")
                                {
                                    int iServerNumber = Convert.ToInt32(voXmlReader.Value);
                                    iCurrentServer = iServerNumber - 1; // ServerNumer starts at 1

                                    // set the property
                                    roQueueDump.Queue.Index[iCurrentIndex].Server[iCurrentServer].Number = voXmlReader.Value;
                                }
                                break;
                            //     status
                            case "status":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Status = voXmlReader.Value;
                                break;
                            //     points?
                            case "points":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Points = voXmlReader.Value;
                                break;
                            //     points_per_hour?
                            case "points_per_hour":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].PointsPerHour = voXmlReader.Value;
                                break;
                            //     progress_rate?
                            case "progress_rate":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].ProgressRate = voXmlReader.Value;
                                break;
                            //     progress?
                            case "progress":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Progress = voXmlReader.Value;
                                break;
                            //     server*
                            //      number                              
                            //      address
                            case "address":
                                //set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Server[iCurrentServer].Address = voXmlReader.Value;
                                break;
                            //     project_number?
                            case "project_number":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].ProjectNumber = voXmlReader.Value;
                                break;
                            //     project_name?
                            case "project_name":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].ProjectName = voXmlReader.Value;
                                break;
                            //     genome*
                            //      unit?
                            case "unit":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Genome.Unit = voXmlReader.Value;
                                break;
                            //      benchmark?
                            case "benchmark":
                                // used by the tags:
                                //  genome
                                //  folding
                                switch (sParentElement)
                                {
                                    case "genome":
                                        // set the property
                                        roQueueDump.Queue.Index[iCurrentIndex].Genome.Benchmark = voXmlReader.Value;
                                        break;
                                    case "folding":
                                        roQueueDump.Queue.Index[iCurrentIndex].Folding.Benchmark = voXmlReader.Value;
                                        break;
                                }
                                break;
                            //      misc?
                            case "misc":
                                // used by the tags:
                                //  genome
                                //  folding
                                switch (sParentElement)
                                {
                                    case "genome":
                                        // set the property
                                        roQueueDump.Queue.Index[iCurrentIndex].Genome.Misc = voXmlReader.Value;
                                        break;
                                    case "folding":
                                        roQueueDump.Queue.Index[iCurrentIndex].Folding.Misc = voXmlReader.Value;
                                        break;
                                }
                                break;
                            //      other?
                            case "other":
                                // used by the tags:
                                //  genome
                                //  folding
                                switch (sParentElement)
                                {
                                    case "genome":
                                        // set the property
                                        roQueueDump.Queue.Index[iCurrentIndex].Genome.Other = voXmlReader.Value;
                                        break;
                                    case "folding":
                                        roQueueDump.Queue.Index[iCurrentIndex].Folding.Other = voXmlReader.Value;
                                        break;
                                }
                                break;
                            //     folding*
                            //      run?
                            case "run":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Folding.Run = voXmlReader.Value;
                                break;
                            //      clone?
                            case "clone":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Folding.Clone = voXmlReader.Value;
                                break;
                            //      generation?
                            case "generation":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Folding.Generation = voXmlReader.Value;
                                break;
                            //      benchmark?
                            //      misc?
                            //      other?
                            //     issue_timestamp?
                            case "issue_timestamp":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].IssueTimestamp = voXmlReader.Value;
                                break;
                            //     begin_timestamp?
                            case "begin_timestamp":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].BeginTimestamp = voXmlReader.Value;
                                break;
                            //     end_timestamp?
                            case "end_timestamp":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].EndTimestamp = voXmlReader.Value;
                                break;
                            //     expect_timestamp?
                            case "expect_timestamp":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].ExpectTimestamp = voXmlReader.Value;
                                break;
                            //     due_timestamp?
                            case "due_timestamp":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].DueTimestamp = voXmlReader.Value;
                                break;
                            //     expire_timestamp?
                            case "expire_timestamp":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].ExpireTimestamp = voXmlReader.Value;
                                break;
                            //     deadline?
                            case "deadline":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Deadline = voXmlReader.Value;
                                break;
                            //     preferred_timestamp?,
                            case "preferred_timestamp":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].PreferredTimestamp = voXmlReader.Value;
                                break;
                            //     preferred_deadline?,
                            case "preferred_deadline":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].PreferredDeadline = voXmlReader.Value;
                                break;
                            //     core?
                            //      url?
                            case "url":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Core.URL = voXmlReader.Value;
                                break;
                            //      version?
                            //     cpu?
                            //      cosm_type?
                            case "cosm_type":
                                // used by the tags:
                                //  cpu
                                //  os
                                switch (sParentElement)
                                {
                                    case "cpu":
                                        // set the property
                                        roQueueDump.Queue.Index[iCurrentIndex].CPU.CosmType = voXmlReader.Value;
                                        break;
                                    case "os":
                                        // set the property
                                        roQueueDump.Queue.Index[iCurrentIndex].OS.CosmType = voXmlReader.Value;
                                        break;
                                }
                                break;
                            //      cosm_subtype?
                            case "cosm_subtype":
                                // used by the tags:
                                //  cpu
                                //  os
                                switch (sParentElement)
                                {
                                    case "cpu":
                                        // set the property
                                        roQueueDump.Queue.Index[iCurrentIndex].CPU.CosmSubtype = voXmlReader.Value;
                                        break;
                                    case "os":
                                        // set the property
                                        roQueueDump.Queue.Index[iCurrentIndex].OS.CosmSubtype = voXmlReader.Value;
                                        break;
                                }
                                break;
                            //      name?
                            case "name":
                                // used by the tags:
                                //  cpu
                                //  os
                                switch (sParentElement)
                                {
                                    case "cpu":
                                        // set the property
                                        roQueueDump.Queue.Index[iCurrentIndex].CPU.Name = voXmlReader.Value;
                                        break;
                                    case "os":
                                        // set the property
                                        roQueueDump.Queue.Index[iCurrentIndex].OS.Name = voXmlReader.Value;
                                        break;
                                }
                                break;
                            //     os?
                            //      cosm_type?
                            //      cosm_subtype?
                            //      name?
                            //     smp_cores?,
                            case "smp_cores":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].SMPCores = voXmlReader.Value;
                                break;
                            //     tag?,
                            case "tag":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Tag = voXmlReader.Value;
                                break;
                            //     passkey?,
                            case "passkey":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Passkey = voXmlReader.Value;
                                break;
                            //     memory?,
                            case "memory":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Memory = voXmlReader.Value;
                                break;
                            //     assignment_info?
                            //      byte_order?
                            case "byte_order":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].AssignmentInfo.ByteOrder = voXmlReader.Value;
                                break;
                            //      timestamp?
                            case "timestamp":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].AssignmentInfo.Timestamp = voXmlReader.Value;
                                break;
                            //      checksum?
                            case "checksum":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].AssignmentInfo.Checksum = voXmlReader.Value;
                                break;
                            //     collection_server?
                            case "collection_server":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].CollectionServer = voXmlReader.Value;
                                break;
                            //     download_timestamp?
                            case "download_timestamp":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].DownloadTimestamp = voXmlReader.Value;
                                break;
                            //     upload_failures?
                            case "upload_failures":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].UploadFailures = voXmlReader.Value;
                                break;
                            //     packet_size_limit?
                            case "packet_size_limit":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].PacketSizeLimit = voXmlReader.Value;
                                break;
                            //     tag?
                            case "qgen":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Qgen = voXmlReader.Value;
                                break;
                            //     username?
                            case "username":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Username = voXmlReader.Value;
                                break;
                            //     team?
                            case "team":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].Team = voXmlReader.Value;
                                break;
                            //     id?
                            case "id":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].ID = voXmlReader.Value;
                                break;
                            //     machine_id?
                            case "machine_id":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].MachineID = voXmlReader.Value;
                                break;
                            //     work_unit?
                            //      file
                            case "file":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].WorkUnit.File = voXmlReader.Value;
                                break;
                            //      size
                            case "size":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].WorkUnit.Size = voXmlReader.Value;
                                break;
                            //      type?
                            case "type":
                                // set the property
                                roQueueDump.Queue.Index[iCurrentIndex].WorkUnit.Type = voXmlReader.Value;
                                break;
                            //    results_successfully_sent?,
                            case "results_successfully_sent":
                                // set the property
                                roQueueDump.Queue.ResultsSuccessfullySent = voXmlReader.Value;
                                break;
                            //    download_rate?
                            //     speed
                            case "speed":
                                // used by the tags:
                                //  download_rate
                                //  upload_rate
                                switch (sParentElement)
                                {
                                    case "download_rate":
                                        // set the property
                                        roQueueDump.Queue.DownloadRate.Speed = voXmlReader.Value;
                                        break;
                                    case "upload_rate":
                                        // set the property
                                        roQueueDump.Queue.UploadRate.Speed = voXmlReader.Value;
                                        break;
                                }
                                break;
                            //     units
                            case "units":
                                // used by the tags:
                                //  download_rate
                                //  upload_rate
                                //  performance_fraction
                                switch (sParentElement)
                                {
                                    case "download_rate":
                                        // set the property
                                        roQueueDump.Queue.DownloadRate.Units = voXmlReader.Value;
                                        break;
                                    case "upload_rate":
                                        // set the property
                                        roQueueDump.Queue.UploadRate.Units = voXmlReader.Value;
                                        break;
                                    case "performance_fraction":
                                        // set the property
                                        roQueueDump.Queue.PerformanceFraction.Units = voXmlReader.Value;
                                        break;
                                }
                                break;
                            //    upload_rate?
                            //     speed
                            //     units      
                            //    performance_fraction?
                            //     fraction
                            case "fraction":
                                // set the property
                                roQueueDump.Queue.PerformanceFraction.Fraction = voXmlReader.Value;
                                break;
                            //     units
                            //    pph?
                            case "pph":
                                // set the property
                                roQueueDump.Queue.PPH = voXmlReader.Value;
                                break;
                            //    ppd?
                            case "ppd":
                                // set the property
                                roQueueDump.Queue.PPD = voXmlReader.Value;
                                break;
                            //    ppw?
                            case "ppw":
                                // set the property
                                roQueueDump.Queue.PPW = voXmlReader.Value;
                                break;
                            //    ppy?
                            case "ppy":
                                // set the property
                                roQueueDump.Queue.PPY = voXmlReader.Value;
                                break;
                        }
                        break;
                }
            }
            voXmlReader.Close();

            // Check whether the document is valid or invalid.
            if (isValid)
            {
                Debug.WriteLine("Document is valid");
            }
            else
            {
                Debug.WriteLine("Document is invalid");
                throw new Exception("Queue Dump XML file is invalid.");
            }

            #endregion Read XML

            return roQueueDump;
        }
    }
}
